@@ -2,11 +2,11 @@ import os, logging, traceback
 from struct import unpack, pack
 import numpy as np
 from aug_sfutils import sfmap, sfobj, str_byt
-from aug_sfutils.sfmap import olbl, ostruc, oattr, header_sfmt
+from aug_sfutils.sfmap import oid, olbl, ostruc, oattr, header_sfmt
 
 fmt = logging.Formatter('%(asctime)s | %(name)s | %(levelname)s: %(message)s', '%H:%M:%S')
 
-logger = logging.getLogger('aug_sfutils.read')
+logger = logging.getLogger('aug_sfutils.shotfile')
 
 if len(logger.handlers) == 0:
     hnd = logging.StreamHandler()
@@ -22,6 +22,7 @@ LOGICAL  = sfmap.typeMap('descr', 'SFfmt', 'LOGICAL')
 
 def getChunk(fname, start, length):
     """Reads the requested byteblock from the binary shotfile"""
+
     rdata = None
     with open(fname, 'rb') as f:
         f.seek(start)
@@ -30,6 +31,7 @@ def getChunk(fname, start, length):
 
 
 def param_length(param):
+    """Determines the byte length of a Parameter"""
 
     parmlen = 16 # ParName(=key), unit, dfmt, n_items
     dfmt = param.dataFormat
@@ -40,6 +42,7 @@ def param_length(param):
 
 
 def parset_length(pset_d):
+    """Determines the byte length of a ParameterSet"""
 
     psetlen = 0
     for param in pset_d.values():
@@ -49,6 +52,7 @@ def parset_length(pset_d):
 
 
 def par2byt(pname, param):
+    """Converts the parameter data into bytes"""
 
     dfmt = param.dataFormat
     n_items = param.n_items
@@ -84,65 +88,55 @@ def par2byt(pname, param):
     return byt
 
 
-
 class SHOTFILE(dict):
 
 
-    def __init__(self, sfpath, sfobj=None):
+    def __init__(self, sfpath, sffull=None):
 
         self.__dict__ = self
 
-        if not os.path.isfile(sfpath):
+        if not os.path.isfile(sfpath) and sffull is None:
             logger.error('Shotfile %s not found', sfpath)
             return None
 
-        self.properties = type('', (), {})()
-        self.properties.path = sfpath
-
-        if sfobj is None: #read automatically
+        if sffull is None: # read shotfile if no SFobj is input
+            self.shotfilePath = sfpath
             self.read_sfh()
             self.set_attributes()
-        else: # for SF writin
-            for key, val in sfobj.__dict__.items():
-                setattr(self, key, val)
-            self.sfout = sfpath
+        else: # for SF writing
+            if isinstance(sffull, dict):
+                for key, val in sffull.items():
+                    setattr(self, key, val)
+            else:
+                for key, val in sffull.__dict__.items():
+                    setattr(self, key, val)
+            self.shotfilePath = sfpath
+            self.sfh2byte() # converts to bytes string (Shotfile content)
 
 
     def read_sfh(self, psets=True):
         """Reads a full shotfile header, including the data of ParamSets, Devices, Lists"""
 
-        sfile = self.properties.path
-
-        self.properties.lists   = []
-        self.properties.parsets = []
-        self.properties.objects = []
-        self.properties.objid   = []
-        self.properties.SFobjects = []
-        self.properties.addrlen = 1
+        self.SFobjects = []
+        self.addressMultiplier   = 1
 
         n_max = 1000
         n_obj = n_max
-        self.properties.related = []
+        self.objId2objName = []
         for j in range(n_max):
-            sfo = SF_OBJECT(j, self.properties)
+            sfo = SF_OBJECT(jobj=j, sfname=self.shotfilePath)
             if hasattr(sfo, 'objectType'):
                 sfo.objid = j
                 onam = str_byt.to_str(sfo.objectName.strip())
                 if sfo.SFOtypeLabel == 'Diagnostic':
-                    self.properties.shot = sfo.shot_nr
                     if n_obj == n_max: # There might be several diags objects in a SFH
+                        self.shotNumber = sfo.shot
+                        self.diagName   = str_byt.to_str(sfo.diag.strip())
                         n_obj = sfo.num_objs
-                self.properties.related.append(onam)
-                self.properties.SFobjects.append(sfo)
-                self.properties.objid.append(j)
-                if sfo.SFOtypeLabel in ('ParamSet', 'Device'):
-                    self.properties.parsets.append(onam)
-                elif sfo.SFOtypeLabel == 'List':
-                    self.properties.lists.append(onam)
-                elif sfo.SFOtypeLabel in ('SignalGroup', 'Signal', 'TimeBase', 'AreaBase'):
-                    self.properties.objects.append(onam)
-                elif sfo.SFOtypeLabel == 'ADDRLEN':
-                    self.properties.addrlen = sfmap.addrsizes[sfo.addrlen]
+                self.objId2objName.append(onam)
+                self.SFobjects.append(sfo)
+                if sfo.SFOtypeLabel == 'ADDRLEN':
+                    self.addressMultiplier = sfmap.addrsizes[sfo.addrlen]
                 self.__dict__[onam] = sfo
             if j >= n_obj - 1:
                 break
@@ -151,17 +145,19 @@ class SHOTFILE(dict):
     def set_attributes(self):
         """Sets useful context info for the entire shotfile, plus data for Lists, ParmSets, Devices"""
 
-        for sfo in self.properties.SFobjects:
-            sfo.address *= self.properties.addrlen
-            sfo.relations  = [self.properties.related[jid]   for jid in sfo.rel if jid != 65535]
-            sfo.relobjects = [self.properties.SFobjects[jid] for jid in sfo.rel if jid != 65535]
+        for sfo in self.SFobjects:
+            sfo.address *= self.addressMultiplier
+            sfo.relations  = [self.objId2objName[jid] for jid in sfo.rel if jid != 65535]
+            sfo.relobjects = [self.SFobjects[jid]     for jid in sfo.rel if jid != 65535]
             if hasattr(sfo, 'dataFormat'):
                 if sfo.dataFormat in sfmap.dtf['SFfmt'].values:
                     sfo.dataType = sfmap.typeMap('SFfmt', 'descr', sfo.dataFormat)
             if sfo.SFOtypeLabel == 'List':
                 sfo.getData()
-                sfo.data = [self.properties.related[jid] for jid in sfo.data]
-
+                sfo.data = []
+                for jid in sfo.data:
+                    sfo.data.append(self.objId2objName[jid])
+ 
             elif sfo.SFOtypeLabel in ('Device', 'ParamSet'):
                 sfo.getData()
                 if sfo.SFOtypeLabel == 'Device':
@@ -191,78 +187,132 @@ class SHOTFILE(dict):
 #----------------
 # Writing section
 #----------------
-    def write_sfh(self):
-# Write SFH file
+    def sfh2byte(self):
+        """Converts all SF objects' metadata to bytes (shotfile header content)"""
+
+# Fill missing attributes in signle objects
+
+        if not hasattr(self, 'SFobjects'):
+            self.SFobjects = []
+            for lbl, sfo in vars(self).items():
+                if hasattr(sfo, 'data') or hasattr(sfo, 'SFOtypeLabel'):
+                    sfo.objectName = lbl
+                    sfo.setDefault()
+                    self.SFobjects.append(sfo)
+
+        self.setSIGNALS()
+        objNames = [sfo.objectName for sfo in self.SFobjects]
+        if 'SIGNALS' not in objNames:
+            self.SFobjects.insert(1, self.SIGNALS)
+        self.setSIGNALSdata()
+
+        objName2objId = {}
+        for jid, sfo in enumerate(self.SFobjects):
+            objName2objId[sfo.objectName] = jid
+
+        for jid, sfo in enumerate(self.SFobjects):
+            if not hasattr(sfo, 'rel'):
+                if hasattr(sfo, 'relations'):
+                    sfo.rel = [objName2objId[oname] for oname in sfo.relations]
+                    while len(sfo.rel) < 8:
+                        sfo.rel.append(65535)
+                else:
+                    sfo.rel = 8*[65535] # defaulting to empty
 
         self.set_length_address()
 
-        f = open(self.sfout, 'wb')
-        for sfo in self.properties.SFobjects:
+        num_objs = len(self.SFobjects)
+        self.SFbytes = b''
+        for sfo in self.SFobjects:
+            if sfo.SFOtypeLabel == 'Diagnostic':
+                sfo.num_objs = num_objs
 # Encode all attributes into byte strings(128)
-            sfo.to_byte()
-            f.write(sfo.sfh_byte)
+            self.SFbytes += sfo.sfoh2byte()
 
 # Write SIGNALS list
-
-        sigs = self.SIGNALS
-        f.write(sigs.objlist)
+        self.SFbytes += self.SIGNALS.data
 
 # Write content of ParSets
-        for sfo in self.properties.SFobjects:
+        for sfo in self.SFobjects:                
             if sfo.SFOtypeLabel == 'ParamSet':
                 pset2byt = b''
                 for pname, param in sfo.data.items():
                     pset2byt += par2byt(pname, param)
-                f.seek(sfo.address) # Ensure proper localisation
-                f.write(pset2byt)
-        f.close()
-        logger.info('Stored binary %s' %self.sfout)
+                self.SFbytes.ljust(sfo.address) # Ensure proper localisation
+                self.SFbytes += pset2byt
 
 
-    def set_signals(self):
+    def putData(self):
+        """Add data to Sf objects"""
+
+        for sfo in self.SFobjects:
+            if sfo.SFOtypeLabel in sfmap.DataObjects:
+                if hasattr(sfo, 'data'):
+                    sfmt  = sfmap.typeMap('SFfmt', 'struc', sfo.dataFormat)
+                    sfo.data = sfo.data.astype('>%s' %sfmt)
+                    if sfo.data.nbytes == sfo.length:
+                        self.SFbytes = self.SFbytes.ljust(sfo.address)
+                        self.SFbytes += sfo.data.tobytes(order='F')
+                    else:
+                        logger.error('Shape/type of %s does not match buffer length', sfo.objectName)
+                        return
+                else:
+                    logger.warning('Missing data for object %s', sfo.objectName)
+
+
+    def dumpShotfile(self):
+        """Dumps Shotfile(header)"""
+
+        sfout = self.shotfilePath
+        if hasattr(self, 'SFbytes'):
+            with open(sfout, 'wb') as f:
+                f.write(self.SFbytes)
+            logger.info('Stored binary %s' %sfout)
+
+
+    def setSIGNALS(self):
+        """Generates 'SIGNALS' list automatically"""
 
 # SIGNALS list generated automatically, override input entry
-        self.sigs = type('', (), {})()
-        self.sigs.nitems = len(self.properties.objects)
-        self.sigs.address = len(self.properties.SFobjects)*128
-        self.sigs.length = self.sigs.nitems*2
-        self.sigs.dataFormat = 3
-        sfmt = sfmap.typeMap('SFfmt', 'struc', self.sigs.dataFormat)
-        objid = [sfo.objid for sfo in self.properties.SFobjects if sfo.objectName in self.properties.objects]
-        byt_objlist = pack('>%d%s' %(self.sigs.nitems, sfmt), *objid)
-        bytlen = len(byt_objlist)
-        dj0 = str_byt.next8(bytlen)
-        self.sigs.objlist = bytearray(dj0)
-        self.sigs.objlist[:bytlen] = byt_objlist
+        self.SIGNALS = SF_OBJECT()
+        self.SIGNALS.dataFormat = sfmap.typeMap('descr', 'SFfmt', 'SHORT_INT')
+        self.SIGNALS.SFOtypeLabel = 'List'
+        self.SIGNALS.objectName = 'SIGNALS'
+
+
+    def setSIGNALSdata(self):
+
+        sfmt = sfmap.typeMap('SFfmt', 'struc', self.SIGNALS.dataFormat)
+        objid = [jid for jid, sfo in enumerate(self.SFobjects) if sfo.SFOtypeLabel in sfmap.DataObjects]
+        self.SIGNALS.nitems = len(objid)
+        self.SIGNALS.data = pack('>%d%s' %(self.SIGNALS.nitems, sfmt), *objid)
+        self.SIGNALS.length = self.SIGNALS.nitems*2
+        self.SIGNALS.address = len(self.SFobjects)*128
+        self.SIGNALS.setDefault()
 
 
     def set_length_address(self):
+        """Recalculates the address and length of SF objects, recreating the 'SIGNALS' list consistently"""
 
 # ParSets
 
         len_psets = 0
-        for sfo in self.properties.SFobjects: # sequence not important
+        for jid, sfo in enumerate(self.SFobjects): # sequence not important
             if sfo.SFOtypeLabel == 'ParamSet':
                 sfo.length = parset_length(sfo.data)
                 len_psets += sfo.length
 
 # Set lengths and addresses
 
-        self.set_signals()
-
-        addr_diag = self.sigs.address + self.sigs.length + len_psets
-        par_addr  = self.sigs.address + self.sigs.length 
+        addr_diag = self.SIGNALS.address + self.SIGNALS.length + len_psets
+        par_addr  = self.SIGNALS.address + self.SIGNALS.length 
         addr_diag = str_byt.next8(addr_diag)
         par_addr  = str_byt.next8(par_addr)
 
         addr = addr_diag
 
-        for sfo in self.properties.SFobjects:
-
-            objectName = sfo.objectName
-            SFOlbl     = sfo.SFOtypeLabel
-            addr_in    = sfo.address # For debugging
-            len_in     = sfo.length  # For debugging
+        for sfo in self.SFobjects:
+            SFOlbl = sfo.SFOtypeLabel
             if hasattr(sfo, 'dataFormat'):
                 type_len = sfmap.typeLength(sfo.dataFormat)
             else:
@@ -270,41 +320,41 @@ class SHOTFILE(dict):
 
             if SFOlbl == 'Diagnostic':
                 sfodiag = sfo
-                sfo.length = sfo.length # GIT: evaluated at the end of the loop
                 sfo.address = addr
             elif SFOlbl == 'List':
-                if objectName == 'SIGNALS':
-                    for key, val in self.sigs.__dict__.items():
+                if sfo.objectName == 'SIGNALS':
+                    for key, val in self.SIGNALS.__dict__.items():
                         setattr(sfo, key, val)
                     addr = addr_diag
             elif SFOlbl in ('Device', 'ParamSet'):
                 sfo.address = par_addr
                 par_addr += sfo.length
-            elif SFOlbl in ('Signal', 'TimeBase', 'SignalGroup', 'AreaBase'):
+            elif SFOlbl in sfmap.DataObjects:
                 sfo.length = sfmap.objectLength(sfo)
                 sfo.address = addr
                 addr += str_byt.next8(sfo.length)
             else:
                 continue
 
-            logger.debug('Address in, out: %d %d', addr_in, sfo.address)
-            logger.debug('Length  in, out: %d %d', len_in , sfo.length )
+        sfodiag.length = addr + sfo.length + addr_diag
 
-        len_tot = addr + sfo.length + addr_diag
-        len_in = sfodiag.length
-        sfodiag.length = len_tot
-        logger.debug('Length  in, out: %d %d', len_in , sfodiag.length )
 
 
 class SF_OBJECT:
-    """Reads the metadata of a generic SF object (sfo) from the SFH's 128byte string.
-    For data, call getData()"""
+    """Reads/writes the metadata of a generic single object (sfo) of a Shotfile from/to the SFH's 128byte string.
+    To fetch the corresponding data, call getData()"""
 
 
-    def __init__(self, jobj, properties):
+    def __init__(self, jobj=None, sfname=None):
+
+
+        if sfname is not None:
+            self.sfname = sfname
+            self.read_sfoh(jobj)
+
+    
+    def read_sfoh(self, jobj):
 
-        self.properties = properties
-        self.sfname = self.properties.path
         byte_str = getChunk(self.sfname, jobj*128, 128)
 
         objnam, self.objectType, self.level, self.status, self.errcode, *rel, \
@@ -337,16 +387,16 @@ class SF_OBJECT:
             self.calibration_type = sfmap.calibLabel[self.cal_type]
         elif SFOlbl in ('SignalGroup', 'Signal'):
             self.index = [self.index1, self.index2, self.index3, self.index4]
-            if self.physunit in sfmap.unit_d.keys():
-                self.phys_unit = sfmap.unit_d[self.physunit]
+            if self.physunit in sfmap.unitLabel.keys():
+                self.phys_unit = sfmap.unitLabel[self.physunit]
             else:
                 logger.warning('No phys. unit found for object %s, key=%d', self.objectName, self.physunit)
                 self.phys_unit = ''
         elif SFOlbl == 'TimeBase':
-            self.timebase_type = sfmap.timebaseLabel[self.tbase_type]
+            self.timebase_type = sfmap.tbaseLabel[self.tbase_type]
         elif SFOlbl == 'AreaBase':
             self.physunit = [self.physunit1, self.physunit2, self.physunit3]
-            self.phys_unit = [sfmap.unit_d[x] for x in self.physunit]
+            self.phys_unit = [sfmap.unitLabel[x] for x in self.physunit]
             self.sizes = [self.size_x, self.size_y, self.size_z]
 
 
@@ -357,7 +407,7 @@ class SF_OBJECT:
             self.getParamSet()
         elif self.SFOtypeLabel == 'List':
             self.getList()
-        elif self.SFOtypeLabel in ('TimeBase', 'AreaBase', 'Signal', 'SignalGroup'):
+        elif self.SFOtypeLabel in sfmap.DataObjects:
             self.getObject(nbeg=nbeg, nend=nend)
 
 
@@ -382,8 +432,8 @@ class SF_OBJECT:
             pname = str_byt.to_str(buf[j0: j0+8])
             meta = type('', (), {})()
             meta.physunit, dfmt, n_items, meta.status = unpack('>4h', buf[j0+8:  j0+16])
-            if meta.physunit in sfmap.unit_d.keys():
-                meta.phys_unit = sfmap.unit_d[meta.physunit]
+            if meta.physunit in sfmap.unitLabel.keys():
+                meta.phys_unit = sfmap.unitLabel[meta.physunit]
             meta.n_items = n_items
             meta.dataFormat = dfmt
 
@@ -408,7 +458,7 @@ class SF_OBJECT:
                 bytlen = val_len * np.dtype(sfmt).itemsize
                 if n_items >= 0:
                     if dfmt == LOGICAL: # Logical, bug if n_items > 1?
-                        meta.dmin, meta.dmax, data = unpack('?2x?1x?', buf[j0: j0+6])
+                        meta.dmin, meta.dmax, data = unpack('>?2x?1x?', buf[j0: j0+6])
                     else:
                         data = np.ndarray((val_len, ), '>%s' %sfmt, buf[j0: j0+bytlen], order='F').copy()
                         meta.dmin = data[0]
@@ -433,7 +483,7 @@ class SF_OBJECT:
                return # do not re-read object if data are there already
         self.nbeg = nbeg
         self.nend = nend
-        if self.SFOtypeLabel in ('SignalGroup', 'Signal', 'TimeBase', 'AreaBase'):
+        if self.SFOtypeLabel in sfmap.DataObjects:
             shape_arr = sfmap.arrayShape(self)
         else:
             logger.error('Object %s is no signal, signalgroup, timebase nor areabase, skipping')
@@ -441,7 +491,7 @@ class SF_OBJECT:
 
         dfmt = self.dataFormat
         if self.SFOtypeLabel == 'TimeBase' and self.length == 0:
-            if self.tbase_type == sfmap.TimebaseType.PPG_prog: # e.g. END:T-LM_END
+            if self.tbase_type == sfmap.tbaseType['PPG_prog']: # e.g. END:T-LM_END
                 self.ppg_time()
             else:   # ADC_intern, e.g. DCN:T-ADC-SL
                 self.data = (np.arange(self.n_steps, dtype=np.float32) - self.n_pre)/self.s_rate
@@ -502,19 +552,121 @@ class SF_OBJECT:
             return (self.time_dim == self.num_dims-1)
 
 
-    def to_byte(self):
+    def sfoh2byte(self):
+        """Converts SF object attributes into a 128bytes header string"""
 
+# val string
         SFOlbl = self.SFOtypeLabel
         sfmt = ostruc[SFOlbl]
         SFOlist = []
         for SFOattr in sfmap.oattr[SFOlbl]:
-            SFOlist.append(getattr(self, SFOattr))
+            if hasattr(self, SFOattr):
+                SFOlist.append(getattr(self, SFOattr))
+            else:
+# String attributes
+                if SFOattr in ('diag', 'hostname', 'date'):
+                    SFOlist.append(b'')
+# (Various) integer types
+                else:
+                    SFOlist.append(0)
         val = pack(sfmt, *SFOlist)
 
+        descr = str_byt.to_byt(self.descr)
         objectName = str_byt.to_byt(self.objectName)
-        descr      = str_byt.to_byt(self.descr)
 
-        self.sfh_byte = pack(header_sfmt, objectName.ljust(8), self.objectType, \
+# Pack all SFh attributes
+        sfoh_byte = pack(header_sfmt, objectName.ljust(8), self.objectType, \
             self.level, self.status, self.errcode, *self.rel, self.address, \
             self.length, val, descr.ljust(64))
-        
+
+        return sfoh_byte
+
+
+    def setDefault(self):
+        """Defaulting missing attributes whenever possible"""
+
+        if not hasattr(self, 'objectName'):
+            logger.error('Missing attribute objectName')
+            return
+
+# Some defaults in case of missing attributes
+
+        type_err = 'SF object type not understood'
+        if not hasattr(self, 'SFOtypeLabel'):
+            if hasattr(self, 'objectType'):
+                self.SFOtypeLabel = olbl[self.objectType]
+            else:
+                if not hasattr(self, 'data'):
+                    logger.error(type_err)
+                    return
+                if self.data.ndim == 1:
+                    self.SFOtypeLabel = 'Signal'      # default: Signal for 1D arrays
+                elif self.data.ndim > 1:
+                    self.SFOtypeLabel = 'SignalGroup' # default: SignalGroup for 2-3D arrays
+                else:
+                    logger.error(type_err)
+                    return
+
+        if not hasattr(self, 'objectType'):
+            self.objectType = oid[self.SFOtypeLabel]
+
+        if self.SFOtypeLabel == 'Diagnostic':
+            self.diag      = str_byt.to_byt(self.objectName).ljust(4)
+            self.version   = 4
+            self.exp       = sfmap.expType['PRIV']
+            self.diag_type = sfmap.diagType['DataAcqu']
+            self.up_limit  = 256
+            self.s_type    = sfmap.shotType['NormalShot']
+        elif self.SFOtypeLabel == 'TimeBase':
+            self.n_steps = len(self.data)
+        elif self.SFOtypeLabel == 'AreaBase':
+            ashape = self.data.shape
+            adim = self.data.ndim
+            self.size_y = 0
+            self.size_z = 0
+            if adim == 1:
+                self.size_x = ashape[0]
+            elif adim == 2:
+                self.size_x, self.size_y = ashape
+            elif adim == 3:
+                self.size_x, self.size_y, self.size_z = ashape
+            if not hasattr(self, 'n_steps'):
+                self.n_steps = 1
+        elif self.SFOtypeLabel in ('Signal', 'SignalGroup'):
+            self.num_dims = self.data.ndim
+            self.index4 = self.data.shape[0]
+            if self.data.ndim > 1:
+                self.index3 = self.data.shape[1]
+            else:
+                self.index3 = 1
+            if self.data.ndim > 2:
+                self.index2 = self.data.shape[2]
+            else:
+                self.index2 = 1
+            if self.data.ndim > 3:
+                self.index1 = self.data.shape[3]
+            else:
+                self.index1 = 1
+            self.index = [self.index1, self.index2, self.index3, self.index4]
+            if not hasattr(self, 'physunit'):
+                if hasattr(self, 'phys_unit'):
+                    self.physunit = sfmap.unitType[self.phys_unit]
+                elif hasattr(self, 'unit'):
+                    self.physunit = sfmap.unitType[self.unit]
+                else:
+                    self.physunit = 0
+
+        if self.SFOtypeLabel in sfmap.DataObjects:
+            if not hasattr(self, 'dataFormat'):
+                if hasattr(self, 'data'):
+                    self.dataFormat = sfmap.typeMap('np', 'SFfmt', self.data.dtype)
+                else:
+                    logger.error('Missing dataFormat for object %s', self.objectName)
+                    return
+
+        if not hasattr(self, 'descr'):
+            self.descr = b''
+
+        for attr in ('level', 'status', 'errcode'):
+            if not hasattr(self, attr):
+                setattr(self, attr, 0)
