@@ -4,7 +4,7 @@ import os, datetime, warnings, logging
 import ctypes as ct
 import numpy as np
 from aug_sfutils import sfmap, str_byt
-from aug_sfutils.sfmap import ObjectID as oid
+from aug_sfutils.sfmap import oid
 
 logger = logging.getLogger('aug_sfutils.ww')
 #logger.setLevel(logging.INFO)
@@ -170,11 +170,11 @@ class WW:
         if hasattr(self, '_diaref'):
             info  = self._wwoinfo(signame)
             otype = info['otype']
-            if otype in (oid.Signal, oid.TimeBase):
+            if otype in (oid['Signal'], oid['TimeBase']):
                 self.SetSignal(signame, data)
-            elif otype == oid.SignalGroup:
+            elif otype == oid['SignalGroup']:
                 self.SetSignalGroup(signame, data)
-            elif otype == oid.AreaBase:
+            elif otype == oid['AreaBase']:
                 self.SetAreabase(signame, data, tim_axis=tim_axis)
 
 
@@ -226,11 +226,11 @@ class WW:
                 _lbuf = ct.byref(lbuf)
                 lsbuf = ct.c_uint64(len(sbuf)) 
                 buffer = ct.c_char_p(sbuf)
-                if otype == oid.Signal:
+                if otype == oid['Signal']:
                     result = libddww.wwsignal_( \
                              _error, self._diaref, name, _type, _lbuf, buffer, \
                              _stride, lname, lsbuf)
-                elif otype == oid.SignalGroup:
+                elif otype == oid['SignalGroup']:
                     result = libddww.wwinsert_( \
                              _error, self._diaref, name, _type, _lbuf, buffer, \
                              _stride, _indices, lname, lsbuf)
@@ -238,23 +238,23 @@ class WW:
                 ctyp = sfmap.typeMap('SFfmt', 'ct', dfmt)
                 data = np.atleast_1d(data)
                 data = np.array(data, dtype=sfmap.typeMap('SFfmt', 'np', dfmt))
-                if otype == oid.Signal:
+                if otype == oid['Signal']:
                     result = libddww.wwsignal_( \
                              _error, self._diaref, name, _type, _lbuf, \
                              data.ctypes.data_as(ct.POINTER(ctyp)), \
                              _stride, lname)
-                elif otype == oid.TimeBase:
+                elif otype == oid['TimeBase']:
                     result = libddww.wwtbase_( \
                              _error, self._diaref, name, _type, _lbuf, \
                              data.ctypes.data_as(ct.POINTER(ctyp)), \
                              _stride, lname)
-                elif otype == oid.SignalGroup:
+                elif otype == oid['SignalGroup']:
                     result = libddww.wwinsert_( \
                              _error, self._diaref, name, _type, _lbuf, \
                              data.ctypes.data_as(ct.POINTER(ctyp)), \
                              _stride, _indices, lname)
 
-            if otype in (oid.Signal, oid.TimeBase):
+            if otype in (oid['Signal'], oid['TimeBase']):
                 sign = signame.decode('utf8')
                 if getError( error ):
                     raise Exception('ww: Error Writing %s' %sign)
@@ -462,11 +462,11 @@ def write_sf(nshot, data_d, sfhdir, diag, exp='AUGD'):
     otype = {}
     for obj, data in data_d.items():
         otype[obj] = ww._wwoinfo(obj)['otype']
-        if otype[obj] == oid.TimeBase:
+        if otype[obj] == oid['TimeBase']:
             ww.SetObject(obj, data)
 
     for obj, data in data_d.items():
-        if otype[obj] == oid.AreaBase:
+        if otype[obj] == oid['AreaBase']:
             if data.ndim == 1:
                 tim_ax = None
             else:
@@ -474,11 +474,11 @@ def write_sf(nshot, data_d, sfhdir, diag, exp='AUGD'):
             ww.SetObject(obj, data, tim_axis=tim_ax)
 
     for obj, data in data_d.items():
-        if otype[obj] == oid.ParamSet:
+        if otype[obj] == oid['ParamSet']:
             logger.info('Writing PS %s', obj)
             for pn, pdata in data.items():
                 ww.SetParameter(obj, pn, pdata)
-        if otype[obj] in (oid.Signal, oid.SignalGroup):
+        if otype[obj] in (oid['Signal'], oid['SignalGroup']):
             ww.SetObject(obj, data)
 
     ww.Close()
